/* File name     : int.c
 * Creation date : August 1995
 * Author        : Y.Yasu, Online group, Physics department, KEK
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/map.h>
#include <sys/buf.h>
#include <sys/vm.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/devio.h>
#include <hal/cpuconf.h>
#include <sys/exec.h>
#include <io/common/devdriver.h>
#include <io/common/handler.h>
#include <io/dec/vme/vbareg.h>
#include <sys/sysconfig.h>
#include <io/dec/tc/tc.h>
#include <machine/cpu.h>

#include <sys/signal.h>
#include <sys/vmmac.h>
#include <sys/mman.h>

#include "int.h"

#define EACCFAULT 200          /* Access violation */
#define ENOMAPREG 201          /* nothing mapping registers */

#define PZERO 25
#define INTERRUPT_PRI ( (PZERO + 1) | PCATCH )

int intprobe(), intintra(), inttimeout();

extern int hz;        /* clock ticks par second */

#define DEBUG         /** select debug mode **/
/*#define DEBUG_LIST*/         /** select debug mode **/

/*
** Data Structure for Device Driver
*/
struct controller *intdinfo[1];
struct driver intdriver = { 
              intprobe,                       /* probe */
              0,                             /* slave */
              0,                             /* cattach */
              0,                             /* dattach */
              0,                             /* go */
              0,                             /* addr_list */
              0,                             /* dev_name */
              0,                             /* dev_list */
              "int",                         /* ctlr_name */
              intdinfo,                      /* ctlr_list */
              0,                             /* xclu */
              sizeof(4),                     /* addr1_size */
              VME_A16_UDATA_D16|VME_BS_BYTE, /* addr1_atype */
              0,                             /* addr2_size */
              0,                             /* addr2_atype */
              0,                             /* ctlr_unattach */
              0                              /* dev_unattach */
      };

int int_interrupt_vector[VME_TOTAL_VECTOR]; /* vector number */
int int_interrupt_channel[VME_TOTAL_VECTOR];/* used for semaphore */
int int_interrupt_enable[VME_TOTAL_VECTOR]; /* enable/disable flag */
int int_interrupt_flag[VME_TOTAL_VECTOR];   /* interrupt waiting flag */
int int_interrupt_count[VME_TOTAL_VECTOR];  /* interrupt counter */

/* for VME word swaps */
union swaps {
    int   longs;
    short shorts[2];
} swaps;

/*
**      Internal Data
*/

static ihandler_id_t * cc_id_t[1][1]; /* Tags for handler_del */

/*
** intprobe() : Autoconfiguration Support Section
*/
intprobe( reg1, ctlr )
io_handle_t reg1;             /* CSR 1 */
struct controller  *ctlr;  
{
    register ihandler_t  handler;
    register struct  vme_handler_info  info;

    /* Sets the info data structure */
    /* Registers the interrupt handlers by setting up the interrupt
       handler structure                                           */
    info.gen_intr_info.configuration_st = (caddr_t)ctlr;
    info.gen_intr_info.intr             = intintra;
    info.gen_intr_info.param            = (caddr_t)0;
    info.gen_intr_info.config_type      = CONTROLLER_CONFIG_TYPE;
    info.vec = ctlr->ivnum;
    info.irq = ctlr->bus_priority;

    /* Sets the handler data structure */
    handler.ih_bus      = ctlr->bus_hd;
    handler.ih_bus_info = (char *)&info;

    cc_id_t[0][0] = handler_add( &handler );
    if( cc_id_t[0][0] == (ihandler_id_t *)NULL ) {
        return( 0 );
    }
    if( handler_enable( cc_id_t[0][0] ) != 0 ) {
        handler_del( cc_id_t[0][0] );
        return( 0 );
    }

    return (4);
}

/*
** intopen()
*/
intopen( dev, flags ,format )
dev_t dev;
int   flags; 
int   format;
{
    register struct controller *ctlr = intdinfo[0];

     /* no such device or address */
    if( ctlr == 0 ) return (ENXIO);
    if( ctlr->alive == 0 ) return (ENXIO);

    return 0;
}

/*
** intclose()
*/
intclose( dev, flags, format )
dev_t  dev;
int    flags;
int    format;
{

    return ( 0 );

}

/*
** intread
*/
intread( dev, uio, flag )
dev_t  dev;
struct uio *uio; 
int    flag;
{
    return ( 0 );
}

/*
** intwrite
*/
intwrite( dev, uio, flag )
dev_t  dev;
struct uio *uio; 
int    flag;
{
    return ( 0 );
}

/*
** intintr() : Interrupt Section
*/
intinit() {
    int i;

    for( i = 0; i < VME_TOTAL_VECTOR; i++)
        int_interrupt_vector[i] = 0;
    for( i = 0; i < VME_TOTAL_VECTOR; i++)
        int_interrupt_channel[i] = 0;
    for( i = 0; i < VME_TOTAL_VECTOR; i++)
        int_interrupt_enable[i] = 0;
    for( i = 0; i < VME_TOTAL_VECTOR; i++)
        int_interrupt_flag[i] = 0;
    for( i = 0; i < VME_TOTAL_VECTOR; i++)
        int_interrupt_count[i] = 0;
  }

intintra( unit )
int unit; {
    intintr(VME_INTR0);
  }

intintrb( unit )
int unit; {
    intintr(VME_INTR1);
  }

intintrc( unit )
int unit; {
    intintr(VME_INTR2);
  }

intintrd( unit )
int unit; {
    intintr(VME_INTR3);
  }

intintre( unit )
int unit; {
    intintr(VME_INTR4);
  }

intintrf( unit )
int unit; {
    intintr(VME_INTR5);
  }

intintrg( unit )
int unit; {
    intintr(VME_INTR6);
  }

intintrh( unit )
int unit; {
    intintr(VME_INTR7);
  }

intintri( unit )
int unit; {
    intintr(VME_INTR8);
  }

intintrj( unit )
int unit; {
    intintr(VME_INTR9);
  }

int intintr( intrno )
int intrno; {

#ifdef DEBUG
    printf("intintr : interrupt occurred. intrno = %d\n", intrno);
#endif
    if( int_interrupt_vector[intrno] && int_interrupt_enable[intrno] &&
        int_interrupt_flag[intrno] ) {
        int_interrupt_count[intrno]++;
        untimeout(inttimeout, intrno);
        wakeup(&int_interrupt_channel[intrno]);
      }
    if( int_interrupt_vector[intrno] && int_interrupt_enable[intrno] &&
        !int_interrupt_flag[intrno] ) {
        int_interrupt_count[intrno]++;
      }
  }


inttimeout(intrno)
int intrno; {

    int_interrupt_flag[intrno] = -1; /* indicate timeout */
    wakeup(&int_interrupt_channel[intrno]);
}

/*
** intioclt : General Purpos I/O Control
*/
intioctl( dev, cmd, data, flag )
dev_t dev;
int cmd;
int *data;
int flag; 
{
    int intrno, vectorno;

    switch (cmd) {
    case VMEIOC_INIT:
        intinit();
        break;
      case VMEIOC_OPEN:
        vectorno = data[0];
#ifdef DEBUG
        printf("vector = %x\n", vectorno);
#endif
        if( vectorno < VME_MIN_VECTOR || vectorno > VME_MAX_VECTOR )
            return (EINVAL);
        intrno = vectorno - VME_MIN_VECTOR;
        int_interrupt_vector[intrno] = vectorno;
#ifdef DEBUG
        printf("int_interrupt_vector[intrno] = %x\n",
               int_interrupt_vector[intrno]);
#endif
        int_interrupt_enable[intrno] = 0;
        int_interrupt_flag[intrno] = 0;
        int_interrupt_count[intrno] = 0;
        data[1] = intrno;
        break;
      case VMEIOC_CLOSE:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        intrno = data[0];
        int_interrupt_vector[intrno] = 0;
        break;
      case VMEIOC_ENABLE:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        if( !int_interrupt_vector[data[0]] )
            return (ENODEV);
        intrno = data[0];
        int_interrupt_enable[intrno] = 1;
        int_interrupt_flag[intrno] = 0;
        break;
      case VMEIOC_DISABLE:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        if( !int_interrupt_vector[data[0]] )
            return (ENODEV);
        intrno = data[0];
        int_interrupt_enable[intrno] = 0;
        int_interrupt_flag[intrno] = 0;
        break;
      case VMEIOC_WAIT:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        if( !int_interrupt_vector[data[0]] )
            return (ENODEV);
        if( !int_interrupt_enable[data[0]] )
            return (ENODEV);
        intrno = data[0];
        spl7();
        /*flag =  -2: signal, -1:timeout, 1: success, 0:already interrupted */
        if ( int_interrupt_flag[intrno] == 1 ) {
#ifdef DEBUG
            printf("VMEIOC_WAIT: signal occurred.\n");
#endif
            untimeout(inttimeout, intrno);
            data[2] = int_interrupt_count[intrno];
            data[3] = -2;
            int_interrupt_flag[intrno] = 0;
            int_interrupt_count[intrno] = 0;
            return EINTR;
	  }
        if( !int_interrupt_count[intrno] ) {
            if( data[1] == 0 )
                timeout(inttimeout, intrno, VME_K_TIMEOUT*hz);
            else
                timeout(inttimeout, intrno, data[1]*hz);
#ifdef DEBUG
                printf("wait_lam: now wait...\n");
#endif
            int_interrupt_flag[intrno] = 1;
            sleep(&int_interrupt_channel[intrno], INTERRUPT_PRI );
	  }
        data[2] = int_interrupt_count[intrno];
        data[3] = int_interrupt_flag[intrno];
#ifdef DEBUG
        printf("int_interrupt_count[intrno] = %d\n", int_interrupt_count[intrno]);
        printf("int_interrupt_flag[intrno] = %d\n", int_interrupt_flag[intrno]);
#endif
        int_interrupt_count[intrno] = 0;
        if( int_interrupt_flag[intrno] == -1 ) {
            int_interrupt_flag[intrno] = 0;
#ifdef DEBUG
            printf("VMEIOC_WAIT: timeout occurred.\n");
#endif
            return ETIMEDOUT;
	  }
        int_interrupt_flag[intrno] = 0;
        break;
      case VMEIOC_NOWAIT:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        if( !int_interrupt_vector[data[0]] )
            return (ENODEV);
        if( !int_interrupt_enable[data[0]] )
            return (ENODEV);
        intrno = data[0];
        data[2] = int_interrupt_count[intrno];
        data[3] = int_interrupt_flag[intrno];
        if ( int_interrupt_count[intrno] ) {
            int_interrupt_count[intrno] = 0;
            int_interrupt_flag[intrno] = 0;
	  }
        break;
      case VMEIOC_CLEAR:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        if( !int_interrupt_vector[data[0]] )
            return (ENODEV);
        if( !int_interrupt_enable[data[0]] )
            return (ENODEV);
        intrno = data[0];
        int_interrupt_flag[intrno] = 0;
        int_interrupt_count[intrno] = 0;
        break;
      case VMEIOC_CHECK:
#ifdef DEBUG
        printf("intrno = %d\n", data[0]);
#endif
        if( data[0] < VME_INTR0 || data[0] > VME_INTR9 )
            return (EINVAL);
        if( !int_interrupt_vector[data[0]] )
            return (ENODEV);
        if( !int_interrupt_enable[data[0]] )
            return (ENODEV);
        intrno = data[0];
        data[2] = int_interrupt_count[intrno];
        data[3] = int_interrupt_flag[intrno];
        break;
    }
    return 0;
}
