/*
 * HISTORY
 */
/*
 *------------------------------------------------------------------------
 *	Source Code for lhdriver, an example VMEbus mmap driver.
 *
 * 	This driver is provided in order for the customer to see how
 *	one might implement mmap() functionality in a VMEbus driver, in
 *	order to access memory mapped to the VMEbus from a user mode 
 *	program.  It was not created to support any particular VME device,
 *	or to provide anything but the minimal code necessary to demonstrate
 *	the mmap() mechanism.
 *
 * 	The driver cur_handlely supports the following platforms:
 *		Alpha AXPvme 160
 *		Alpha AXPvme 64
 *	
 *------------------------------------------------------------------------
 */

/*********************************************************
		INCLUDE FILES
*********************************************************/

/* Header files required by memory mapped device driver */
#include "lh.h" /* Driver header file generated by config */  
#include "lhdriver.h" /* device specific header */
#include <sys/param.h>
#include <sys/buf.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <io/dec/vme/vbareg.h> 
#include <sys/vmmac.h>
#include <sys/user.h>
#include <sys/file.h>
#include <sys/map.h>
#include <machine/cpu.h>
#include <sys/uio.h>
#include <hal/cpuconf.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <io/common/devdriver.h>
#include <kern/task.h>
#include <sys/malloc.h>
#define lh_MAXPHYS (256*1024)

#define DEBUG 
#ifdef DEBUG 
#define Cprintf printf
#else
#define Cprintf ;
#endif

struct buf lhbuf[NLH];
/***********************************************************
		DECLARATIONS
***********************************************************/

/* Driver routines declarations */
int lh_mmap();
int lh_probe(), lh_ioctl();
int lh_read(), lh_write();
int lh_open(), lh_close();
int lh_minphys(),lh_strategy();

/* Array of pointers to controller structures */ 
struct controller *lh_cinfo[NLH]; 

/* Declare and initialize lhdriver structure */ 
struct driver lhdriver = {lh_probe,0,0,0,0,0,0,0,"lh",lh_cinfo,0,0,0,0,0,0,0};

/* Define a softc structure for use by the driver routines */
struct lh_softc{
	io_handle_t   lh_iohandle; /* csr iohandle for PIO READ/WRITE   */
        unsigned long lh_pio_addr; /* VME address for pio transfers     */
	unsigned long lh_pio_size; /* Size to map for pio transfers     */
	unsigned long lh_pio_am;  /* Address modifier of lh_pio_addr   */
	unsigned int  lh_pio_access; /* Access type for lh_pio_addr     */
	unsigned int  lh_pio_is_mapped; /* Set if lh_pio_addr is mapped */
        int 	      opened;       /* Set if device cur_handlely opened    */
	int           dma_flags;   /* Flags for DMA */
	vme_atype_t   mode;        /* read/write mode(DMA/NON_DMA)*/
	vme_addr_t    vmeaddr;	   /* VME Address for DMA */
	struct buf    *bp;
	dma_handle_t      sc_dma_handle;
} lh_softc[NLH]; 



/***********************************************************
		AUTOCONFIGURATION
***********************************************************/

/***********************************************************
		Probe Routine
***********************************************************/
lh_probe(caddr_t reg, struct controller *ctlr)
{
        register int unit = ctlr->ctlr_num;

	/* Save the ctlr info */
	lh_cinfo[unit] = ctlr;
        return (1); 
}

/***********************************************************
		Open Routine
***********************************************************/
lh_open(dev_t dev, int flag, int format)
{
	int unit = minor(dev);		/* Get device unit number */
        register struct controller *ctlr = lh_cinfo[unit];

	if (unit > NLH)
		return(ENXIO);		/* Invalid unit number */

	if(lh_softc[unit].opened)	/* Only one at a time please */
		return(EBUSY);

	lh_softc[unit].opened = 1; 	/* All ok, indicate device opened */

	/* Zero out the softc structure.  */
	lh_softc[unit].lh_iohandle = 0L;
	lh_softc[unit].lh_pio_addr = 0L;
	lh_softc[unit].lh_pio_size = 0L;
	lh_softc[unit].lh_pio_am   = 0L;
	lh_softc[unit].lh_pio_access = 0L;
	lh_softc[unit].lh_pio_is_mapped = 0L;
	lh_softc[unit].dma_flags = 0;
	lh_softc[unit].mode = NON_DMA;
 	lh_softc[unit].vmeaddr = 0L;	

	return (0); 
}


/***********************************************************
		Close Routine
***********************************************************/
lh_close(dev_t dev, int flag, int format)
{
	int unit = minor(dev);	/* Get device unit number */
        register struct controller *ctlr = lh_cinfo[unit];

	lh_softc[unit].opened = 0;	/* Indicate device is closed */

	/* Unmap PIO */
	if (lh_softc[unit].lh_pio_is_mapped)
	    {
	    vba_unmap_csr(ctlr, ctlr->addr);
	    ctlr->addr = (caddr_t)0;
	    lh_softc[unit].lh_iohandle = 0L;
	    lh_softc[unit].lh_pio_addr = 0L;
	    lh_softc[unit].lh_pio_size = 0L;
	    lh_softc[unit].lh_pio_am   = 0L;
	    lh_softc[unit].lh_pio_access = 0L;
	    lh_softc[unit].lh_pio_is_mapped = 0L;
	    lh_softc[unit].dma_flags = 0;
	    lh_softc[unit].mode = NON_DMA;
 	    lh_softc[unit].vmeaddr = 0L;	
	    }

	return (0); 
}

/***********************************************************
		Ioctl Routine
***********************************************************/
lh_ioctl(dev_t dev, int cmd, caddr_t data, int flag)
{
	int unit = minor(dev);		/* Get device unit number */
        register struct controller *ctlr = lh_cinfo[unit];
	int ret_val = 0;

	switch(cmd) {

	case LH_MAP_PIO_ADDR:	
		/* Map outbound to a user-supplied VME address, size, and 
		 * address modifier, supplied in the data parameter, which
		 * is passed in as an pio_info structure type.
		 */
		if (lh_softc[unit].lh_pio_is_mapped)
		    {
                    printf("lh_ioctl: LH_MAP_PIO_ADDR failed, a PIO mapping \n");
		    printf("	      already exists.  You must unmap the old \n");
		    printf("	      mapping first. \n");
		    ret_val = -1;
		    }
		if ( lh_softc[unit].mode == DMA ){
		    printf("lh_ioctl: LH_SET_DMA failed, a PIO mapping \n") ;
                    printf("      set already.  You must unset the old \n ");
                    printf("      dma first. \n");
                    ret_val = -1;
                    }
		    
		/* Perform the outbound mapping */
		ctlr->addr = (caddr_t)0;
		ctlr->addr = (caddr_t) vba_map_csr(
				ctlr,
				((struct pio_info *)data)->pio_addr,
				((struct pio_info *)data)->pio_size,
				((struct pio_info *)data)->pio_am);

		if (ctlr->addr == NULL)
		    {
		    printf("lh_ioctl: vba_map_csr failed, pio_addr = 0x%x\n",
				((struct pio_info *)data)->pio_addr);
		    printf("				  pio_size = 0x%x\n",
				((struct pio_info *)data)->pio_size);
		    printf("				  pio_am   = 0x%x\n",
				((struct pio_info *)data)->pio_am);
		    printf("				  pio_access = 0x%x\n",
				((struct pio_info *)data)->pio_access);
		    ret_val = -1;
		    }

		/* Save the PIO mapping info */
		lh_softc[unit].lh_iohandle = (io_handle_t)ctlr->addr;
		lh_softc[unit].lh_pio_is_mapped = 1;
		lh_softc[unit].lh_pio_addr = ((struct pio_info *)data)->pio_addr;
	    	lh_softc[unit].lh_pio_size = ((struct pio_info *)data)->pio_size;
	    	lh_softc[unit].lh_pio_am   = ((struct pio_info *)data)->pio_am;
	    	lh_softc[unit].lh_pio_access  = ((struct pio_info *)data)->pio_access;
		lh_softc[unit].mode = NON_DMA;

		break;

	case LH_UNMAP_PIO_ADDR:	
		/*  Unmap pio access */
		if (lh_softc[unit].lh_pio_is_mapped)
		    {
		    vba_unmap_csr(ctlr, ctlr->addr);
		    ctlr->addr = (caddr_t)0;
	            lh_softc[unit].lh_iohandle = 0L;
		    lh_softc[unit].lh_pio_is_mapped = 0L;
		    lh_softc[unit].lh_pio_addr = 0L;
	    	    lh_softc[unit].lh_pio_size = 0L;
	    	    lh_softc[unit].lh_pio_am   = 0L;
		    lh_softc[unit].lh_pio_access = 0L;
		    }
		else 
		    {
		    printf("lh_ioctl: LH_UNMAP_PIO_ADDR, there is no cur_handle PIO mapping \n");
  		    ret_val = -1;
		    }
		break;

	case LH_GET_PIO_INFO:   
		/* Ret PIO mapping info to user */
		((struct pio_info *)data)->pio_addr = lh_softc[unit].lh_pio_addr;
		((struct pio_info *)data)->pio_size = lh_softc[unit].lh_pio_size;
		((struct pio_info *)data)->pio_am   = lh_softc[unit].lh_pio_am;
		((struct pio_info *)data)->pio_access = lh_softc[unit].lh_pio_access;
		break;

	case LH_SET_DMA_INFO:
		if ( lh_softc[unit].mode == DMA ){
		    printf("lh_ioctl: LH_SET_DMA failed, a PIO mapping \n") ;
                    printf("      set already.  You must unset the old \n ");
                    printf("      dma first. \n");
                    ret_val = -1;
                    }
		lh_softc[unit].dma_flags = ((struct dma_info *)data)->flag;
		lh_softc[unit].mode = DMA;
	 	lh_softc[unit].vmeaddr = ((struct dma_info *)data)->vmeaddr;	
		break;

	case LH_UNSET_DMA_INFO:
		lh_softc[unit].dma_flags = 0L;
		lh_softc[unit].mode = NON_DMA;
	 	lh_softc[unit].vmeaddr = 0L;	
		break;

	default:
                printf("lh_ioctl: unknown command \n");
		ret_val = 0;
	}

	return (ret_val);
}



int lh_read(dev, uio, flag)
dev_t  dev;      /* Major/minor device number */
struct uio *uio; /* Pointer to uio structure */
int flag;        /* Access mode of device */
{
      int unit = minor(dev);
      int err = 0 ;
      int cnt;
      volatile long tmp;
      io_handle_t  cur_handle = lh_softc[unit].lh_iohandle;
      if ((unit > NLH) || (unit < 0))
             return (EIO);
      Cprintf("request is %d\n",uio->uio_resid);
      if ( lh_softc[unit].mode & NON_DMA ){
	        Cprintf("Now In read NON_DMA\n");
		cur_handle = lh_softc[unit].lh_iohandle;
	        Cprintf("current_handle for NON_DMA %x\n",&cur_handle);
		if( cur_handle == 0 ) return (EIO);
		while((cnt = uio->uio_resid) && (err == 0)) {
			if(cnt >= 8)cnt = 8; 
			if( (cnt >= 4) && (cnt !=8) )cnt = 4;
			if( (cnt >= 2) && (cnt !=8) && (cnt !=4) )
						cnt = 2; 
                        tmp = read_io_port( cur_handle,
                                              cnt,
                                              0); /* Read data
                                                    from VME memory */
			if (tmp < 0){
				printf ("read_io_port error\n");
				return (EIO);
			}
			Cprintf("read data = %s\n",&tmp);
                        err = uiomove(&tmp,cnt,uio);
			cur_handle = cur_handle + cnt;
                }
		Cprintf("Result = %d\n",err);
                return(err);
      }
      if ( lh_softc[unit].mode & DMA ){
	Cprintf("Now In read DMA\n");
      	if ( lh_softc[unit].dma_flags & DMA_IN ){
		return (physio(lh_strategy, &lhbuf[unit], dev, B_READ,
                        lh_minphys, uio));
      	}
      }
      return(err);
}


int lh_write(dev, uio, flag)
dev_t  dev;      /* Major/minor device number */
struct uio *uio; /* Pointer to uio structure */
int flag;        /* Access mode of device */
{

      register int unit = minor(dev);
      int err = 0;
      int cnt,stat;
      long tmp; 
      io_handle_t  cur_handle = lh_softc[unit].lh_iohandle;
      if ((unit > NLH) || (unit < 0))
             return (EIO);
      if ( lh_softc[unit].mode & NON_DMA ){
		Cprintf("Now In WRITE NON_DMA\n");
		cur_handle = lh_softc[unit].lh_iohandle;
	        Cprintf("current_handle for NON_DMA %x\n",&cur_handle);
		if( cur_handle == 0 ) return (EIO);
		while((cnt = uio->uio_resid) && (err == 0)) {
			if(cnt >= 8)cnt = 8; 
			if( (cnt >= 4) && (cnt != 8) )cnt = 4;
			if( (cnt >= 2) && (cnt != 8) && (cnt != 4) )
						cnt = 2; 
			err = uiomove(&tmp,cnt,uio);
			Cprintf("write data = %s\n",&tmp);
			write_io_port( lh_softc[unit].lh_iohandle,
                                              cnt,
                                              0,
                                              tmp); /* Write data
                                                    to VME memory */
			cur_handle = cur_handle + cnt;
                }
		return(err);
      }
      if ( lh_softc[unit].mode & DMA ){
      	Cprintf("Now In WRITE DMA\n");
      	if ( lh_softc[unit].dma_flags & DMA_OUT ){
      		return (physio(lh_strategy, &lhbuf[unit], dev, B_WRITE,
                        lh_minphys, uio));
      	}
      }
      return(err);
}

int lh_minphys (bp)
register struct buf *bp;
{
/***************************************************
 * The following code defines the maximum size of  *
 * the DMA for this device.  If the maximum size   *
 * for the DMA exceeds lh_MAXPHYS:                 *
 *                                                 *
 * Return to physio which will call strategy.      *
 * This can occur several times until transfer is  *
 * complete.                                       *
 ***************************************************/

      if (bp->b_bcount > lh_MAXPHYS)
              bp->b_bcount = lh_MAXPHYS;
      return;
}


/***************************************************
 *     Strategy Section                            *
 *                                                 *
 *----------------- lh_strategy -------------------*
 *                                                 *
 * The lh_strategy interface is called by the      *
 * phsyio kernel interface.  (The physio interface *
 * is called by the lh_read and lh_write           *
 * driver interfaces.)  The lh_strategy            *
 * interface first makes sure that the user buffer *
 * is both readable and writeable.  It determines  *
 * if the buffer size is larger than lh_MAXPHYS    *
 * and then initiates the I/O.                     *
 ***************************************************/

int lh_strategy(bp)
struct buf  *bp; /* Pointer to buf structure */
{

/***************************************************
 * The following code initializes:                 *
 *   o A unit variable to the minor device number  *
 *   o A controller structure to its associated    *
 *     LH device                                   *
 *   o A lh_softc structure to the address of      *
 *     the lh_softc associated with this           *
 *     lh  device                                  *
 *   o A variable to store read, write, and enable *
 *     interrupts status information               *
 *   o An sg_entry structure pointer that the code *
 *     uses in calls to dma_get_curr_sgentry and   *
 *     write_io_port.                              *
 ***************************************************/

      register int unit = minor(bp->b_dev);
      register struct controller *ctlr = lh_cinfo[unit];
      register struct lh_softc *sc = &lh_softc[unit];
      u_int	rtn_flags = 0;
      unsigned long rtn_bc=0;
      unsigned long max_bc = 256*1024;
      char *buffer;
      int 	s;

/***************************************************
 * Check the validity of the unit number and the   *
 * pointers.                                       *
 ***************************************************/

      if ( (!ctlr) || (!sc) )
                return (EINVAL);
       sc->bp = bp;
       sc->sc_dma_handle = (sglist_t) NULL;

       if (lh_softc[unit].dma_flags & DMA_IN){
	    Cprintf("Now In Strategy B_READ\n");
	    /*MALLOC( buffer , char *, max_bc , M_DEVBUF, M_WAITOK );*/
	/* set DMA info */
	    if (sc->vmeaddr==NULL){
	    		bp->b_error = EIO;
	 		iodone(bp);
	     }

            rtn_flags = vba_set_dma_addr(ctlr, sc->dma_flags , sc->vmeaddr);
	    if ( rtn_flags==0 )
	        Cprintf("rtn_flags=0 in set_dma\n");

            rtn_bc = dma_map_alloc( (unsigned long)bp->b_bcount,
                                        ctlr, &sc->sc_dma_handle, rtn_flags);
	    if ( rtn_bc==0 )
	        Cprintf("rtn_bc=0 in map alloc\n");

            rtn_bc = dma_map_load((unsigned long)bp->b_bcount,
                                   (vm_offset_t)bp->b_un.b_addr,
                                   bp->b_proc,
                                   ctlr,
                                   &sc->sc_dma_handle,
                                   0,
                                   rtn_flags);
            /*rtn_bc = dma_map_load( max_bc,
                                   (vm_offset_t)buffer,
                                   0,
                                   ctlr,
                                   &sc->sc_dma_handle,
                                   0,
                                   rtn_flags);*/
	    if ( rtn_bc==0 ){
	        Cprintf("rtn_bc=0 in map load\n");
		bp->b_error = EIO; 
                bp->b_flags |= B_ERROR;                            
                iodone(bp);                                        
                return;                                            
        	}         
	
	/*
	 * Call vba_dma() to startup and monitor AXPvme SBC block-
	 * mode DMA engine. the return value from vba_dma is the
	 * actual bytes transfered.
	 */
	    Cprintf("GO DMA_READ\n");
	    s = splbio();
	    rtn_bc = vba_dma( ctlr, sc->sc_dma_handle );
	    splx(s);

	/*
	 * unload and free DMA resources
	 */
	    Cprintf("READ DATA %s\n",buffer);
	    /*copyout(buffer,bp->b_un.b_addr,bp->b_resid);*/
	    dma_map_unload( rtn_flags, sc->sc_dma_handle );
	    dma_map_dealloc(sc->sc_dma_handle);
            iodone(bp);                                        
	    return(rtn_bc);
	}
	
        if (lh_softc[unit].dma_flags & DMA_OUT){
	/* set DMA info */
	    Cprintf("Now In Strategy B_WRITE\n");
	    if (sc->vmeaddr==NULL){
	    		bp->b_error = EIO;
	 		iodone(bp);
	     }
	    Cprintf("DMA write data= %s\n",bp->b_un.b_addr);

            rtn_flags = vba_set_dma_addr(ctlr, sc->dma_flags , sc->vmeaddr);
	    if ( rtn_flags==0 )
	        Cprintf("rtn_flags=0 in set_dma\n");

            rtn_bc = dma_map_alloc( (unsigned long) bp->b_bcount,
                                        ctlr, &sc->sc_dma_handle, rtn_flags);
	    if ( rtn_bc==0 )
	        Cprintf("rtn_bc=0 in map alloc\n");

            rtn_bc = dma_map_load((unsigned long) bp->b_bcount,
                                  (vm_offset_t)bp->b_un.b_addr,
                                   bp->b_proc,
                                   ctlr,
                                   &sc->sc_dma_handle,
                                   0,
                                   rtn_flags);

	    if ( rtn_bc==0 ){
	        Cprintf("rtn_bc=0 In map_load \n");
		bp->b_error = EIO; 
                bp->b_flags |= B_ERROR;                            
                iodone(bp);                                        
                return;                                            
        	}         

	    Cprintf("GO DMA_WRITE\n");

	    s = splbio();
            rtn_bc  =  vba_dma(ctlr,sc->sc_dma_handle);
	    splx(s);
	/*
	 * unload and free DMA resources
	 */
	    dma_map_unload( rtn_flags, sc->sc_dma_handle );
	    dma_map_dealloc(sc->sc_dma_handle);
            iodone(bp);                                        
	    return(rtn_bc);
	    
	}

        if ( !rtn_bc  )
                return (EINVAL);

	return(rtn_bc);
}

/***********************************************************
		MEMORY MAPPING

   The lh_mmap routine is invoked by the kernel as a
   result of an application calling the mmap(2) system
   call.  lh_mmap returns the page frame number corresponding 
   to the page at the specified offset.
***********************************************************/

int lh_mmap(dev_t dev, off_t off, int prot)
{
long kpfnum,flags;
register int unit = minor(dev);
register struct controller *ctlrptr = lh_cinfo[unit];
register io_handle_t reg = (io_handle_t) ctlrptr->addr;
caddr_t phys_addr=0;

 	if (lh_softc[unit].lh_pio_is_mapped)
	    {
	    if ((lh_softc[unit].lh_pio_am) & VME_DENSE)
		flags = HANDLE_DENSE_SPACE;
	    else
		flags = HANDLE_SPARSE_SPACE;
 
	    /* OR in the access type : eg.  HANDLE_LONGWORD */
	    flags |= (u_int)(lh_softc[unit].lh_pio_access);

            /* Change csr I/O handle to a physical address */
            phys_addr =
            	((caddr_t)iohandle_to_phys((io_handle_t)reg, flags));

	    Cprintf("*** io_handle = 0x%lx\n",(u_long)reg);
	    Cprintf("*** Bus physical address = 0x%lx\n",phys_addr);

            phys_addr = (char *)PHYS_TO_KSEG(phys_addr);

	    Cprintf("*** Kseg address = 0x%lx\n",phys_addr);

            kpfnum = (long) ((((u_long)phys_addr + off) >> 13) & (0x1fffffL));

	    Cprintf("*** Page frame number = 0x%x\n",kpfnum);

            return(kpfnum);
	    }
	else
	    {
	    printf("lh_mmap: You must do an outbound mapping first. \n");
	    return(-1);
	    }
}
