/*
 * lh_test.c:  program to test sparse and dense mmap capabilities
 * 	       of the lhdriver test driver.  This test and the
 *	       associated driver are ONLY supported on Alpha AXPvme
 *	       systems.
 * 
 * cc lh_test.c -o lh_test
 */
#include <sys/types.h> 
#include <sys/errno.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/ioctl.h>
#include <c_asm.h> 
#include <mach/alpha/boolean.h>
#include <io/common/devdriver.h>  /* Defines HANDLE_LONGWORD */
#include <io/common/handler.h>    /* For use with vbareg.h */
#include <io/dec/vme/vbareg.h>    /* Defined VME_A24, VME_DENSE, etc. */
#include "/sys/io/lh/lhdriver.h"

/***********************************************
** User must modify the following information **
** in order to test the lh mmap driver        **
***********************************************/

/* Define the VME address of the device that you want to map outbound to.
 * NOTE: this VME address MUST already be mapped inbound on the VME
 * by some other device or driver.  If it isn't, then the driver will 
 * generate a VME outbound mapping error from the ioctl mapping function.
 */
#define VME_PIO_ADDRESS 0x100000

/* Define the size in bytes you want to map to */
#define BUF_SIZE 0x100  

/* Define the address modifier value for sparse space mapping.
 * These values are defined in vbareg.h.
 */
#define SPARSE_AM (VME_A32|VME_SDATA|VME_BS_NOSWAP)

/* Define the address modifier value for dense space mapping
 * These values are defined in vbareg.h.
 */
#define DENSE_AM (VME_A32|VME_SDATA|VME_BS_NOSWAP|VME_DENSE)

/* Define the access type for sparse space mapping.  For sparse space,
 * the macros (below) manipulate the address correctly for read/write
 * accesses, so it is not necessary to modify SPARSE_ACCESS value.
 */
#define SPARSE_ACCESS (HANDLE_LONGWORD)

/* Define the access type for dense space mapping.   
 * Only HANDLE_LONGWORD and HANDLE_QUADWORD are valid for dense mappings.  
 * In addition, for Alpha AXPvme platforms, HANDLE_LONGWORD is treated the 
 * same as HANDLE_QUADWORD, thus it is not necessary to modify DENSE_ACCESS 
 * value.
 */
#define DENSE_ACCESS (HANDLE_LONGWORD)
/***********************************************************************/

int fd;
caddr_t status;
int *bufp;      /* address of start of buffer in mapped area */
int *mapped_va; /* address of start of mapped area */
int *va;
u_long offset_from_base = 0;
extern int errno;

/* Define a MACRO to calculate the offset into the address returned
 * from mmap(), at which the specified VME address is mapped -- ie
 * in order to find the exact virtual address of where the VME address
 * is mapped to, the user must add the calculated offset to the mmap()
 * return address.
 *
 * For Dense space mappings, the offset into the mmap()'d address
 * is exactly the VME address MODULO one alpha page (8192 dec, ie 0x2000).
 * ie for VME_PIO_ADDRESS 0xA40140 you would add 0x140 to the
 * mmap() address to access byte 0 of the mapped VME address area.
 *
 * For Sparse space mappings, the offset into the mmap()'d address
 * is exactly the VME address MODULO 256kb (ie 0x100),
 * the result of which is multiplied by 32 decimal (due to the layout
 * of sparse mappings on an Alpha AXPvme which has one valid byte for
 * every 32 bytes).
 * ie for VME_PIO_ADDRESS 0xA40140 you would add (0x40 * 32) = 0x800
 * to the mmap() address to access byte 0 of the mapped VME address
 * area.
 */
#define get_offset(vme_address,vme_am) \
    ((vme_am) & VME_DENSE) ? ((vme_address) % 0x2000) : (((vme_address) % 0x100) * 32)

/* Define MACROs for manipulating mmap'd memory in sparse space 
 * for the AXPvme platform.
 */
#define read_shift(address,data,width) \
    { data = ( (data >> ((address & 3) << 3)) & \
             ( (1 << ((8*width)-1)) + ((1L << ((8*width)-1L))-1L) ) ); }

#define read_sparse_byte(base, offset, dest) { \
    unsigned char *va; \
    unsigned long data=0; \
    va = (unsigned char *) ((((u_long)offset) << 5) + \
        ((u_long)base )); \
    data = (unsigned long) *(int *)va; \
    read_shift(((u_long)offset & 3), data, 1); \
    dest = (unsigned char) data; }

#define read_sparse_word(base, offset, dest) { \
    unsigned short *va; \
    unsigned long data=0; \
    va = (unsigned short *) ((((u_long)offset) << 5) + \
        ((u_long)base ) | 8); \
    data = (unsigned long) *(int *)va; \
    read_shift(((u_long)offset & 3), data, 2); \
    dest = (unsigned short) data; }

#define read_sparse_int(base, offset, dest) { \
    unsigned int *va; \
    unsigned long data=0; \
    va = (unsigned int *) ((((u_long)offset) << 5) + \
        ((u_long)base ) | 0x18); \
    data = (unsigned long) *(int *)va; \
    read_shift(((u_long)offset & 3), data, 4); \
    dest = (unsigned int) data; }

#define write_sparse_byte(base, offset,data) { \
    unsigned char *va; \
    unsigned long write_data = \
        (unsigned long) data << (((u_long)offset & 3) << 3); \
    va = (unsigned char *) ((((u_long)offset) << 5) + \
        ((u_long)base )); \
    *(int *)va = write_data; \
    (void)asm("mb"); }

#define write_sparse_word(base,offset,data) { \
    unsigned short *va; \
    unsigned long write_data = \
        (unsigned long) data << (((u_long)offset & 3) << 3); \
    va = (unsigned short *) ((((u_long)offset) << 5) + \
        ((u_long)base ) | 8); \
    *(int *)va = write_data; \
    (void)asm("mb");  }

#define write_sparse_int(base, offset, data) { \
    unsigned int *va; \
    unsigned long write_data = \
        (unsigned long) data << (((u_long)offset & 3) << 3); \
    va = (unsigned int *) ((((u_long)offset) << 5) + \
        ((u_long)base ) | 0x18); \
    *(int *)va = write_data; \
    (void)asm("mb"); }

/* pio_info structure is defined in lhdriver.h, and contains
 * fields which get passed in/out of the ioctl function calls
 * of lhdriver.c.
 */
struct pio_info set_lhdata;  
struct pio_info get_lhdata;  
struct dma_info set_lhdma;

#define PAGE_MULT 32    /* For sparse space mapping, one byte every 32 
			 *  bytes is valid.
			 */

main()
{
size_t len;    
unsigned char  *read_byte_data, *write_byte_data, *byte_vaddr;
unsigned short *read_word_data, *write_word_data, *word_vaddr;
unsigned int   *read_int_data,  *write_int_data, *int_vaddr;
unsigned long  *read_quad_data,  *write_quad_data, *quad_vaddr;
int *read_buffer = NULL;
int *write_buffer = NULL;
volatile char *readchar="";
int err,i,j;

/* Allocate buffers to test the mapping */
write_buffer = (int *)malloc(BUF_SIZE);
if(!(write_buffer))
    {
    perror("write malloc");
    exit(1);
    }

read_buffer = (int *)malloc(BUF_SIZE);
if (!(read_buffer))
    {
    perror("read malloc");
    exit(1);
    }

/*****************************************************************
***** First test mapping to the VME address using sparse space  **
******************************************************************/

printf("************** Test Sparse Space Mapping ***********\n");

/* Open the mmap device driver */
if ((fd = open("/dev/lh",O_RDWR)) < 0)
    {
    perror("Error opening lh0");
    exit(1);
    }

printf("Successfully opened /dev/lh0 \n");

/* LH_MAP_PIO_ADDR ioctl function does an outbound mapping to 
 * pio_addr for a size of pio_size and address modifier pio_am.
 * The driver maps outbound on the VME to the exact buffer size, but 
 * the user then mmap's to (32 * buf_size).
 * It doesn't matter what pio_access value is, since the macros
 * do the correct manipulations for sparse space.
 */
len = BUF_SIZE;
set_lhdata.pio_addr   = VME_PIO_ADDRESS;    
set_lhdata.pio_size   = len;  
set_lhdata.pio_am     = SPARSE_AM;
set_lhdata.pio_access = SPARSE_ACCESS;

if ((err = ioctl(fd, LH_MAP_PIO_ADDR, &set_lhdata)) < 0)  
    {
    perror("LH_MAP_PIO_ADDR");
    close(fd);  
    exit(1);
    }
printf("Successfully mapped the VME address to sparse space \n");

if ((err = ioctl(fd, LH_GET_PIO_INFO, &get_lhdata)) < 0)
    {
    perror("LH_GET_PIO_INFO");
    close(fd);
    exit(1);
    }

printf("PIO mapping info: \n");
printf(" VME address = 0x%x\n",get_lhdata.pio_addr);
printf(" Buffer size = 0x%x\n",get_lhdata.pio_size);
printf(" Address modifier = 0x%x\n",get_lhdata.pio_am);
printf(" PIO Access value = 0x%x\n",get_lhdata.pio_access);

/* mmap() will return the starting address of the mapped region, truncated
 * to a page boundary (See lh_mmap(), which returns the page frame number
 * only, and thus we must offset into that page from the user program).  
 * Calculate the offset into that page where the buffer is mapped to.  Use 
 * this offset to ADD to the length of the buffer that we want to mmap(), in 
 * order to map the entire buffer size.
 */
offset_from_base = (u_long)get_offset(VME_PIO_ADDRESS,SPARSE_AM);

printf("*** offset_from_base for VME address 0x%x for sparse space = 0x%lx\n",VME_PIO_ADDRESS,offset_from_base);

/* mmap a size of 32 times the buffer size PLUS the offset into 
 * the mmap'd page (calculated above), for sparse space mapping.
 */
len = (BUF_SIZE * 32) + offset_from_base;
mapped_va = (int *)mmap((caddr_t)0,len,PROT_READ|PROT_WRITE|PROT_EXEC,MAP_SHARED,fd,0);
if (mapped_va == (int *)-1) 
    {
    perror("mmap: failed");
    close(fd);
    exit();
    }
printf("Successfully called mmap \n");

/* Calculate the base address of the buffer, given the offset and the 
 * address returned from mmap.
 */
(u_long)bufp = (u_long)mapped_va + (u_long)offset_from_base;
printf("*** buffer address = 0x%x\n",(u_long)bufp);

printf("Writing byte data\n"); 

write_byte_data = (unsigned char*)write_buffer;
read_byte_data  = (unsigned char*)read_buffer;
byte_vaddr      = (unsigned char*)bufp;
for (i=0; i < BUF_SIZE; i++, byte_vaddr++) 
    {
    write_byte_data[i] = i;
    read_byte_data[i] = 0;

    /* Write to the mmap'd memory using the sparse space macros */
    write_sparse_byte(bufp, ((u_long)byte_vaddr - (u_long)bufp), write_byte_data[i]);
    }

printf("Reading byte data\n"); 

byte_vaddr = (unsigned char*)bufp;
for (i=0; i<BUF_SIZE; i++, byte_vaddr++)
    /* Read the mmap'd memory using the sparse space macros */
    read_sparse_byte(bufp, ((u_long)byte_vaddr - (u_long)bufp),read_byte_data[i]);


printf("Comparing byte data\n"); 

for (i=0, j=0; i<BUF_SIZE; i++) 
    {
    if (write_byte_data[i] != read_byte_data[i]) 
	{
        printf("Compare failed i = 0x%x read = 0x%x, write = 0x%x\n", 
 	    i, read_byte_data[i]), write_byte_data[i];
        if (++j > 32) break; 
        }
    }


printf("Writing word data\n");

write_word_data = (unsigned short*)write_buffer;
read_word_data  = (unsigned short*)read_buffer;
word_vaddr      = (unsigned short*)bufp;
for (i=0; i<BUF_SIZE/2; i++, word_vaddr++) 
    {
    write_word_data[i] = i;
    read_word_data[i]=0;
    write_sparse_word(bufp, ((u_long)word_vaddr - (u_long)bufp), write_word_data[i]);
    }


printf("Reading word data\n"); 

word_vaddr = (unsigned short*)bufp;
for (i=0; i<BUF_SIZE/2; i++, word_vaddr++)
    read_sparse_word(bufp, ((u_long)word_vaddr - (u_long)bufp),read_word_data[i]);


printf("Comparing word data\n");

for (i=0, j=0; i<BUF_SIZE/2; i++) 
    {
    if (write_word_data[i] != read_word_data[i]) 
	{
        printf("Compare failed i = 0x%x read = 0x%x, write = 0x%x\n",
            i, read_word_data[i]), write_word_data[i];
        if (++j > 32) break; 
        }
    }


printf("Writing integer data\n");

write_int_data = (unsigned int*)write_buffer;
read_int_data  = (unsigned int*)read_buffer;
int_vaddr      = (unsigned int*)bufp;

for (i=0; i<BUF_SIZE/4; i++, int_vaddr++) 
    {
    write_int_data[i] = i;
    read_int_data[i]=0;
    write_sparse_int(bufp, ((u_long)int_vaddr - (u_long)bufp), write_int_data[i]);
    }


printf("Reading integer data\n");

int_vaddr = (unsigned int*)bufp;
for (i=0; i<BUF_SIZE/4; i++, int_vaddr++)
    read_sparse_int(bufp, ((u_long)int_vaddr - (u_long)bufp),read_int_data[i]);


printf("Comparing integer data\n");

for (i=0, j=0; i<BUF_SIZE/4; i++) 
    {
    if (write_int_data[i] != read_int_data[i]) 
	{
        printf("Compare failed i = 0x%x read = 0x%x, write = 0x%x\n",
            i, read_int_data[i], write_int_data[i]);
        if (++j > 32) break; 
	}
    }

/* Unmap the current outbound mapping to sparse space */
if ((err = ioctl(fd, LH_UNMAP_PIO_ADDR, &set_lhdata.pio_addr)) < 0)  
    {
    perror("LH_UNMAP_PIO_ADDR");
    close(fd);  
    exit(1);
    }

/* Unmap the user-level mapping */
status = (caddr_t) munmap((caddr_t)mapped_va,len);


/*****************************************************************
***** Now test mapping to the VME address using dense space  *****
******************************************************************/

printf("\n************** Test Dense Space Mapping ***********\n");

/* LH_MAP_PIO_ADDR ioctl function does an outbound mapping to pio_addr for 
 * a size of pio_size and address modifier pio_am.  The driver maps outbound 
 * on the VME to the exact buffer size, and the user mmap's to the exact 
 * buffer size also, since it is mapped to dense space.
 */
len = BUF_SIZE;
set_lhdata.pio_addr   = VME_PIO_ADDRESS;    
set_lhdata.pio_size   = len;  
set_lhdata.pio_am     = DENSE_AM;
set_lhdata.pio_access = DENSE_ACCESS;

if ((err = ioctl(fd, LH_MAP_PIO_ADDR, &set_lhdata)) < 0)  
    {
    perror("LH_MAP_PIO_ADDR");
    close(fd);  
    exit(1);
    }

printf("Successfully mapped the VME address to dense space \n");

if ((err = ioctl(fd, LH_GET_PIO_INFO, &get_lhdata)) < 0)
    {
    perror("LH_GET_PIO_INFO");
    close(fd);
    exit(1);
    }

printf("PIO mapping info: \n");
printf(" VME address = 0x%x\n",get_lhdata.pio_addr);
printf(" Buffer size = 0x%x\n",get_lhdata.pio_size);
printf(" Address modifier = 0x%x\n",get_lhdata.pio_am);
printf(" PIO Access value = 0x%x\n",get_lhdata.pio_access);

/* mmap() will return the starting address of the mapped region, truncated
 * to a page boundary (See lh_mmap(), which returns the page frame number
 * only, and thus we must offset into that page from the user program).  
 * Calculate the offset into that page where the buffer is mapped to.  Use 
 * this offset to ADD to the length of the buffer that we want to mmap(), in 
 * order to map the entire buffer size.
 */
offset_from_base = (u_long)get_offset(VME_PIO_ADDRESS,DENSE_AM);

printf("*** offset_from_base for VME address 0x%x for dense space = 0x%lx\n",VME_PIO_ADDRESS,offset_from_base);

/* mmap to the size of the buffer size PLUS the offset into 
 * the mmap'd page (calculated above), for dense space mapping.
 */
len = BUF_SIZE + offset_from_base;
mapped_va = (int *)mmap((caddr_t)0,len,PROT_READ|PROT_WRITE|PROT_EXEC,MAP_SHARED,fd,0);
if (mapped_va == (int *)-1)
    {
    perror("mmap: failed");
    close(fd);
    exit();
    }
printf("Successfully called mmap \n");

/* Calculate the base address of the buffer, given the offset and the 
 * address returned from mmap.
 */
(u_long)bufp = (u_long)mapped_va + (u_long)offset_from_base;
printf("*** base buffer address = 0x%x\n",(u_long)bufp);


/**** NOTE: Only integer and quadword (32 and 64 bit) accesses can
 **** be done reliable to densely mapped memory.  Byte and word accesses 
 **** will work (on AXPvme systems) but they will do a read followed by a 
 **** write operation on 32 or 64 bits only.
 ****/

printf("Writing integer data\n");

write_int_data = (unsigned int*)write_buffer;
read_int_data  = (unsigned int*)read_buffer;
int_vaddr      = (unsigned int*)bufp;

for (i=0; i < BUF_SIZE/4; i++, int_vaddr++) 
    {
    write_int_data[i] = i;
    read_int_data[i]=0;

    /* Write directly to int_vaddr, since this is dense space */
    *int_vaddr = write_int_data[i];
    }

printf("Reading integer data\n");

int_vaddr = (unsigned int*)bufp;
for (i=0; i < BUF_SIZE/4; i++, int_vaddr++)
    read_int_data[i] = *int_vaddr;


printf("Comparing integer data\n");

for (i=0, j=0; i < BUF_SIZE/4; i++) 
    {
    if (write_int_data[i] != read_int_data[i]) 
	{
        printf("Compare failed i = 0x%x read = 0x%x, write = 0x%x\n",
            i, read_int_data[i], write_int_data[i]);
        if (++j > 32) break; 
	}
    }


printf("Writing quad data\n");

write_quad_data = (unsigned long*)write_buffer;
read_quad_data  = (unsigned long*)read_buffer;
quad_vaddr      = (unsigned long*)bufp;

for (i=0; i < BUF_SIZE/8; i++, quad_vaddr++) 
    {
    write_quad_data[i] = i;
    read_quad_data[i]  = 0;

    /* Write directly to quad_vaddr, since this is dense space */
    *quad_vaddr = write_quad_data[i];
    }

printf("Reading quad data\n");

quad_vaddr = (unsigned long*)bufp;
for (i=0; i < BUF_SIZE/8; i++, quad_vaddr++)
    read_quad_data[i] = *quad_vaddr;


printf("Comparing quad data\n");

for (i=0, j=0; i < BUF_SIZE/8; i++) 
    {
    if (write_quad_data[i] != read_quad_data[i]) 
	{
        printf("Compare failed i = 0x%x read = 0x%lx, write = 0x%lx\n",
            i, read_quad_data[i], write_quad_data[i]);
        if (++j > 32) break; 
	}
    }

/* Unmap the current outbound mapping to sparse space */
  if ((err = ioctl(fd, LH_UNMAP_PIO_ADDR, &set_lhdata.pio_addr)) < 0)  
    {
    perror("LH_UNMAP_PIO_ADDR");
    close(fd);  
    exit(1);
    } 

/* Unmap the user-level mapping */
status = (caddr_t) munmap((caddr_t)mapped_va,len);

/* Close the device file, which will unmap the current outbound mapping  */
close(fd);


}
